'''
Created on 2011-03-27

@author: eposse
'''
import unittest

from libutilitaspy.categories.finite_sets import Set, Function, empty, singleton, equalizer, coequalizer
from libutilitaspy.categories.limits import Cone
from libutilitaspy.categories.colimits import CoCone

class TestFiniteSetCategory(unittest.TestCase):


    def setUp(self):
        pass

    def tearDown(self):
        pass
    
    def test_empty(self):
        Z = empty()
        X = Set(['a','b','c'])
        self.assertEquals(Z.vertex, Set([]))
        X_cocone = CoCone(X,[])
        u = Z.unique_arrow(X_cocone)
        self.assertTrue(isinstance(u, Function))
        self.assertEquals(u.source, Z.vertex)
        self.assertEquals(u.target, X)

    def test_singleton(self):
        S = singleton('*')
        X = Set(['a','b','c'])
        self.assertEquals(S.vertex, Set(['*']))
        X_cone = Cone(X, [])
        u = S.unique_arrow(X_cone)
        self.assertTrue(isinstance(u, Function))
        self.assertEquals(u.source, X)
        self.assertEquals(u.target, S.vertex)
        self.assertEquals(u('a'), '*')
        self.assertEquals(u('b'), '*')
        self.assertEquals(u('c'), '*')
        
    def test_product(self):
        A = Set(['a','b','c'])
        B = Set([1,2])
        AxB = A * B
        self.assertEquals(AxB.vertex, Set([('a',1),('b',1),('c',1),('a',2),('b',2),('c',2)]))
        self.assertEquals(AxB.proj1(('a',1)), 'a')
        self.assertEquals(AxB.proj2(('a',1)), 1)
        C = Set(['x','y'])
        f1 = Function(C, A, {'x': 'a', 'y': 'c'})
        f2 = Function(C, B, {'x': 2, 'y': 1})
        C_cone = Cone(C, [f1, f2])
        u = AxB.unique_arrow(C_cone)
        self.assertEquals(u('x'), ('a',2))
        self.assertEquals(u('y'), ('c',1))
        comp1 = (AxB.proj1 * u)
        comp2 = (AxB.proj2 * u)
        self.assertEquals(f1('x'),comp1('x'))
        self.assertEquals(f1('y'),comp1('y'))
        self.assertEquals(f2('x'),comp2('x'))
        self.assertEquals(f2('y'),comp2('y'))

    def test_coproduct(self):
        A = Set(['a','b','c'])
        B = Set(['b','d'])
        AUB = A + B
        self.assertEquals(AUB.vertex, Set([(0,'a'), (0,'b'), (0,'c'), (1,'b'), (1,'d')]))
        self.assertEquals(AUB.inj1('b'), (0,'b'))
        self.assertEquals(AUB.inj2('b'), (1,'b'))
        C = Set(['x','y'])
        f1 = Function(A, C, {'a':'x','b':'x','c':'y'})
        f2 = Function(B, C, {'b':'y','d':'x'})
        C_cocone = CoCone(C, [f1, f2])
        u = AUB.unique_arrow(C_cocone)
        self.assertEquals(u((0,'b')), 'x')
        self.assertEquals(u((1,'b')), 'y')
        comp1 = (u * AUB.inj1)
        comp2 = (u * AUB.inj2)
        self.assertEquals(f1('a'),comp1('a'))
        self.assertEquals(f1('b'),comp1('b'))
        self.assertEquals(f1('c'),comp1('c'))
        self.assertEquals(f2('b'),comp2('b'))
        self.assertEquals(f2('d'),comp2('d'))
    
    def test_equalizer(self):
        A = Set(['a','b','c'])
        B = Set([1,2])
        f = Function(A, B, {'a':1,'b':2,'c':1})
        g = Function(A, B, {'a':2,'b':2,'c':1})
        Eq = equalizer(f,g)
        self.assertEquals(Eq.vertex, Set(['b','c']))
        self.assertRaises(AssertionError, Eq.eq, ('a'))
        self.assertEquals(Eq.eq('b'), 'b')
        self.assertEquals(Eq.eq('c'), 'c')
        M = Set(['w','x','y','z'])
        m = Function(M, A, {'w':'b','x':'b','y':'a','z':'c'})
        comp1 = f * m
        comp2 = g * m
        self.assertEquals(comp1('w'),comp2('w'))
        self.assertEquals(comp1('x'),comp2('x'))
        self.assertNotEquals(comp1('y'),comp2('y'))
        self.assertEquals(comp1('z'),comp2('z'))
        M_cone = Cone(M, [m])
        u = Eq.unique_arrow(M_cone)
        self.assertEquals(u('w'), 'b')
        self.assertEquals(u('x'), 'b')
        self.assertRaises(AssertionError, u, ('y'))  
        self.assertEquals(u('z'), 'c')
        comp = Eq.eq * u
        self.assertEquals(comp('w'), m('w'))
        self.assertEquals(comp('x'), m('x'))
        self.assertRaises(AssertionError, comp, ('y'))
        self.assertEquals(comp('z'), m('z'))
    
    def test_coequalizer_1(self):
        A = Set(['a1'])
        B = Set(['b1'])
        f = Function(A, B, { 'a1': 'b1' })
        g = Function(A, B, { 'a1': 'b1' })
        part1 = Set(['b1'])
        partition = Set([part1])
        CoEq = coequalizer(f,g)
        self.assertEquals(CoEq.vertex, partition)
        self.assertEquals(CoEq.coeq('b1'), part1)
        M = Set(['x','y'])
        m = Function(B, M, {'b1': 'y'})
        comp1 = m * f
        comp2 = m * g
        self.assertEquals(comp1('a1'),comp2('a1'))
        M_cocone = CoCone(M, [m])
        u = CoEq.unique_arrow(M_cocone)
        self.assertEquals(u(part1), 'y')
        comp = u * CoEq.coeq
        self.assertEquals(comp('b1'), m('b1'))
        
    def test_coequalizer_2(self):
        A = Set(['a1'])
        B = Set(['b1', 'b2'])
        f = Function(A, B, { 'a1': 'b1' })
        g = Function(A, B, { 'a1': 'b2' })
        part1 = Set(['b1', 'b2'])
        partition = Set([part1])
        CoEq = coequalizer(f,g)
        self.assertEquals(CoEq.vertex, partition)
        self.assertEquals(CoEq.coeq('b1'), part1)
        self.assertEquals(CoEq.coeq('b2'), part1)
        M = Set(['x','y'])
        m = Function(B, M, {'b1': 'x', 'b2': 'x'})
        comp1 = m * f
        comp2 = m * g
        self.assertEquals(comp1('a1'),comp2('a1'))
        M_cocone = CoCone(M, [m])
        u = CoEq.unique_arrow(M_cocone)
        self.assertEquals(u(part1), 'x')
        comp = u * CoEq.coeq
        self.assertEquals(comp('b1'), m('b1'))
        self.assertEquals(comp('b2'), m('b2'))
        
    def test_coequalizer_3(self):
        A = Set(['a1','a2'])
        B = Set(['b1','b2'])
        f = Function(A, B, { 'a1': 'b1', 'a2': 'b1' })
        g = Function(A, B, { 'a1': 'b1', 'a2': 'b2' })
        part1 = Set(['b1', 'b2'])
        partition = Set([part1])
        CoEq = coequalizer(f,g)
        self.assertEquals(CoEq.vertex, partition)
        self.assertEquals(CoEq.coeq('b1'), part1)
        self.assertEquals(CoEq.coeq('b2'), part1)
        M = Set(['x','y'])
        m = Function(B, M, {'b1': 'x', 'b2': 'x'})
        comp1 = m * f
        comp2 = m * g
        self.assertEquals(comp1('a1'),comp2('a1'))
        self.assertEquals(comp1('a2'),comp2('a2'))
        M_cocone = CoCone(M, [m])
        u = CoEq.unique_arrow(M_cocone)
        self.assertEquals(u(part1), 'x')
        comp = u * CoEq.coeq
        self.assertEquals(comp('b1'), m('b1'))
        self.assertEquals(comp('b2'), m('b2'))
        
    def test_coequalizer_4(self):
        A = Set(['a1','a2'])
        B = Set(['b1','b2'])
        f = Function(A, B, { 'a1': 'b1', 'a2': 'b2' })
        g = Function(A, B, { 'a1': 'b1', 'a2': 'b2' })
        part1 = Set(['b1'])
        part2 = Set(['b2'])
        partition = Set([part1,part2])
        CoEq = coequalizer(f,g)
        self.assertEquals(CoEq.vertex, partition)
        self.assertEquals(CoEq.coeq('b1'), part1)
        self.assertEquals(CoEq.coeq('b2'), part2)
        M = Set(['x','y'])
        m = Function(B, M, {'b1': 'x', 'b2': 'x'})
        comp1 = m * f
        comp2 = m * g
        self.assertEquals(comp1('a1'),comp2('a1'))
        self.assertEquals(comp1('a2'),comp2('a2'))
        M_cocone = CoCone(M, [m])
        u = CoEq.unique_arrow(M_cocone)
        self.assertEquals(u(part1), 'x')
        self.assertEquals(u(part2), 'x')
        comp = u * CoEq.coeq
        self.assertEquals(comp('b1'), m('b1'))
        self.assertEquals(comp('b2'), m('b2'))
        
    def test_coequalizer_5(self):
        A = Set(['a1','a2'])
        B = Set(['b1','b2'])
        f = Function(A, B, { 'a1': 'b1', 'a2': 'b2' })
        g = Function(A, B, { 'a1': 'b1', 'a2': 'b2' })
        part1 = Set(['b1'])
        part2 = Set(['b2'])
        partition = Set([part1,part2])
        CoEq = coequalizer(f,g)
        self.assertEquals(CoEq.vertex, partition)
        self.assertEquals(CoEq.coeq('b1'), part1)
        self.assertEquals(CoEq.coeq('b2'), part2)
        M = Set(['x','y'])
        m = Function(B, M, {'b1': 'x', 'b2': 'y'})
        comp1 = m * f
        comp2 = m * g
        self.assertEquals(comp1('a1'),comp2('a1'))
        self.assertEquals(comp1('a2'),comp2('a2'))
        M_cocone = CoCone(M, [m])
        u = CoEq.unique_arrow(M_cocone)
        self.assertEquals(u(part1), 'x')
        self.assertEquals(u(part2), 'y')
        comp = u * CoEq.coeq
        self.assertEquals(comp('b1'), m('b1'))
        self.assertEquals(comp('b2'), m('b2'))
        
    def test_coequalizer_6(self):
        A = Set(['a1','a2'])
        B = Set(['b1','b2'])
        f = Function(A, B, { 'a1': 'b1', 'a2': 'b1' })
        g = Function(A, B, { 'a1': 'b2', 'a2': 'b2' })
        part1 = Set(['b1', 'b2'])
        partition = Set([part1])
        CoEq = coequalizer(f,g)
        self.assertEquals(CoEq.vertex, partition)
        self.assertEquals(CoEq.coeq('b1'), part1)
        self.assertEquals(CoEq.coeq('b2'), part1)
        M = Set(['x','y'])
        m = Function(B, M, {'b1': 'x', 'b2': 'y'})
        comp1 = m * f
        comp2 = m * g
        self.assertNotEquals(comp1('a1'),comp2('a1'))
        self.assertNotEquals(comp1('a2'),comp2('a2'))
        M_cocone = CoCone(M, [m])
        u = CoEq.unique_arrow(M_cocone)
        self.assertEquals(u(part1), 'x')
        self.assertNotEquals(u(part1), 'y')
        comp = u * CoEq.coeq
        self.assertEquals(comp('b1'), m('b1'))
        self.assertNotEquals(comp('b2'), m('b2'))
        
    def test_coequalizer_7(self):
        A = Set(['a1','a2'])
        B = Set(['b1','b2'])
        f = Function(A, B, { 'a1': 'b1', 'a2': 'b1' })
        g = Function(A, B, { 'a1': 'b2', 'a2': 'b2' })
        part1 = Set(['b1', 'b2'])
        partition = Set([part1])
        CoEq = coequalizer(f,g)
        self.assertEquals(CoEq.vertex, partition)
        self.assertEquals(CoEq.coeq('b1'), part1)
        self.assertEquals(CoEq.coeq('b2'), part1)
        M = Set(['x','y'])
        m = Function(B, M, {'b1': 'x', 'b2': 'x'})
        comp1 = m * f
        comp2 = m * g
        self.assertEquals(comp1('a1'),comp2('a1'))
        self.assertEquals(comp1('a2'),comp2('a2'))
        M_cocone = CoCone(M, [m])
        u = CoEq.unique_arrow(M_cocone)
        self.assertEquals(u(part1), 'x')
        self.assertEquals(u(part1), 'x')
        comp = u * CoEq.coeq
        self.assertEquals(comp('b1'), m('b1'))
        self.assertEquals(comp('b2'), m('b2'))
        
    def test_coequalizer_8(self):
        A = Set(['a1','a2'])
        B = Set(['b1','b2','b3','b4'])
        f = Function(A, B, { 'a1': 'b1', 'a2': 'b2' })
        g = Function(A, B, { 'a1': 'b3', 'a2': 'b4' })
        part1 = Set(['b1', 'b3'])
        part2 = Set(['b2', 'b4'])
        partition = Set([part1,part2])
        CoEq = coequalizer(f,g)
        self.assertEquals(CoEq.vertex, partition)
        self.assertEquals(CoEq.coeq('b1'), part1)
        self.assertEquals(CoEq.coeq('b2'), part2)
        self.assertEquals(CoEq.coeq('b3'), part1)
        self.assertEquals(CoEq.coeq('b4'), part2)
        M = Set(['x','y'])
        m = Function(B, M, {'b1':'x', 'b2':'y', 'b3':'x', 'b4':'y'})
        comp1 = m * f
        comp2 = m * g
        self.assertEquals(comp1('a1'),comp2('a1'))
        self.assertEquals(comp1('a2'),comp2('a2'))
        M_cocone = CoCone(M, [m])
        u = CoEq.unique_arrow(M_cocone)
        self.assertEquals(u(part1), 'x')
        self.assertEquals(u(part2), 'y')
        comp = u * CoEq.coeq
        self.assertEquals(comp('b1'), m('b1'))
        self.assertEquals(comp('b2'), m('b2'))
        
    def test_coequalizer_9(self):
        A = Set(['a1','a2'])
        B = Set(['b1','b2','b3','b4','b5'])
        f = Function(A, B, { 'a1':'b1', 'a2':'b2', 'a3':'b4' })
        g = Function(A, B, { 'a1':'b2', 'a2':'b3', 'a3':'b4' })
        part1 = Set(['b1', 'b2', 'b3'])
        part2 = Set(['b4'])
        part3 = Set(['b5'])
        partition = Set([part1,part2,part3])
        CoEq = coequalizer(f,g)
        self.assertEquals(CoEq.vertex, partition)
        self.assertEquals(CoEq.coeq('b1'), part1)
        self.assertEquals(CoEq.coeq('b2'), part1)
        self.assertEquals(CoEq.coeq('b3'), part1)
        self.assertEquals(CoEq.coeq('b4'), part2)
        self.assertEquals(CoEq.coeq('b5'), part3)
        M = Set(['x','y'])
        m = Function(B, M, {'b1':'x', 'b2':'x', 'b3':'x', 'b4':'y', 'b5':'y'})
        comp1 = m * f
        comp2 = m * g
        self.assertEquals(comp1('a1'),comp2('a1'))
        self.assertEquals(comp1('a2'),comp2('a2'))
        self.assertEquals(comp1('a3'),comp2('a3'))
        M_cocone = CoCone(M, [m])
        u = CoEq.unique_arrow(M_cocone)
        self.assertEquals(u(part1), 'x')
        self.assertEquals(u(part2), 'y')
        self.assertEquals(u(part3), 'y')
        comp = u * CoEq.coeq
        self.assertEquals(comp('b1'), m('b1'))
        self.assertEquals(comp('b2'), m('b2'))
        self.assertEquals(comp('b3'), m('b3'))
        self.assertEquals(comp('b4'), m('b4'))
        self.assertEquals(comp('b5'), m('b5'))
        
    def test_coequalizer_10(self):
        A = Set(['a1','a2', 'a3', 'a4'])
        B = Set(['b1','b2','b3','b4','b5'])
        f = Function(A, B, { 'a1':'b1', 'a2':'b2', 'a3':'b4', 'a4': 'b3' })
        g = Function(A, B, { 'a1':'b2', 'a2':'b3', 'a3':'b4', 'a4': 'b5' })
        part1 = Set(['b1', 'b2', 'b3', 'b5'])
        part2 = Set(['b4'])
        partition = Set([part1,part2])
        CoEq = coequalizer(f,g)
        self.assertEquals(CoEq.vertex, partition)
        self.assertEquals(CoEq.coeq('b1'), part1)
        self.assertEquals(CoEq.coeq('b2'), part1)
        self.assertEquals(CoEq.coeq('b3'), part1)
        self.assertEquals(CoEq.coeq('b4'), part2)
        self.assertEquals(CoEq.coeq('b5'), part1)
        M = Set(['x','y'])
        m = Function(B, M, {'b1':'x', 'b2':'x', 'b3':'x', 'b4':'y', 'b5':'x'})
        comp1 = m * f
        comp2 = m * g
        self.assertEquals(comp1('a1'),comp2('a1'))
        self.assertEquals(comp1('a2'),comp2('a2'))
        self.assertEquals(comp1('a3'),comp2('a3'))
        self.assertEquals(comp1('a4'),comp2('a4'))
        M_cocone = CoCone(M, [m])
        u = CoEq.unique_arrow(M_cocone)
        self.assertEquals(u(part1), 'x')
        self.assertEquals(u(part2), 'y')
        comp = u * CoEq.coeq
        self.assertEquals(comp('b1'), m('b1'))
        self.assertEquals(comp('b2'), m('b2'))
        self.assertEquals(comp('b3'), m('b3'))
        self.assertEquals(comp('b4'), m('b4'))
        self.assertEquals(comp('b5'), m('b5'))

    def test_coequalizer_11(self):
        A = Set(['a1','a2', 'a3', 'a4'])
        B = Set(['b1','b2','b3','b4','b5'])
        f = Function(A, B, { 'a1':'b1', 'a2':'b2', 'a3':'b4', 'a4': 'b4' })
        g = Function(A, B, { 'a1':'b2', 'a2':'b3', 'a3':'b4', 'a4': 'b5' })
        part1 = Set(['b1', 'b2', 'b3'])
        part2 = Set(['b4', 'b5'])
        partition = Set([part1,part2])
        CoEq = coequalizer(f,g)
        self.assertEquals(CoEq.vertex, partition)
        self.assertEquals(CoEq.coeq('b1'), part1)
        self.assertEquals(CoEq.coeq('b2'), part1)
        self.assertEquals(CoEq.coeq('b3'), part1)
        self.assertEquals(CoEq.coeq('b4'), part2)
        self.assertEquals(CoEq.coeq('b5'), part2)
        M = Set(['x','y'])
        m = Function(B, M, {'b1':'x', 'b2':'x', 'b3':'x', 'b4':'y', 'b5':'y'})
        comp1 = m * f
        comp2 = m * g
        self.assertEquals(comp1('a1'),comp2('a1'))
        self.assertEquals(comp1('a2'),comp2('a2'))
        self.assertEquals(comp1('a3'),comp2('a3'))
        self.assertEquals(comp1('a4'),comp2('a4'))
        M_cocone = CoCone(M, [m])
        u = CoEq.unique_arrow(M_cocone)
        self.assertEquals(u(part1), 'x')
        self.assertEquals(u(part2), 'y')
        comp = u * CoEq.coeq
        self.assertEquals(comp('b1'), m('b1'))
        self.assertEquals(comp('b2'), m('b2'))
        self.assertEquals(comp('b3'), m('b3'))
        self.assertEquals(comp('b4'), m('b4'))
        self.assertEquals(comp('b5'), m('b5'))
        
    def test_coequalizer_12(self):
        A = Set(['a1','a2', 'a3', 'a4'])
        B = Set(['b1','b2','b3','b4','b5','b6','b7'])
        f = Function(A, B, { 'a1':'b1', 'a2':'b2', 'a3':'b4', 'a4': 'b4' })
        g = Function(A, B, { 'a1':'b2', 'a2':'b3', 'a3':'b4', 'a4': 'b5' })
        part1 = Set(['b1', 'b2', 'b3'])
        part2 = Set(['b4', 'b5'])
        part3 = Set(['b6'])
        part4 = Set(['b7'])
        partition = Set([part1,part2,part3,part4])
        CoEq = coequalizer(f,g)
        self.assertEquals(CoEq.vertex, partition)
        self.assertEquals(CoEq.coeq('b1'), part1)
        self.assertEquals(CoEq.coeq('b2'), part1)
        self.assertEquals(CoEq.coeq('b3'), part1)
        self.assertEquals(CoEq.coeq('b4'), part2)
        self.assertEquals(CoEq.coeq('b5'), part2)
        self.assertEquals(CoEq.coeq('b6'), part3)
        self.assertEquals(CoEq.coeq('b7'), part4)
        M = Set(['x','y'])
        m = Function(B, M, {'b1':'x', 'b2':'x', 'b3':'x', 'b4':'y', 'b5':'y', 'b6':'x', 'b7':'y'})
        comp1 = m * f
        comp2 = m * g
        self.assertEquals(comp1('a1'),comp2('a1'))
        self.assertEquals(comp1('a2'),comp2('a2'))
        self.assertEquals(comp1('a3'),comp2('a3'))
        self.assertEquals(comp1('a4'),comp2('a4'))
        M_cocone = CoCone(M, [m])
        u = CoEq.unique_arrow(M_cocone)
        self.assertEquals(u(part1), 'x')
        self.assertEquals(u(part2), 'y')
        self.assertEquals(u(part3), 'x')
        self.assertEquals(u(part4), 'y')
        comp = u * CoEq.coeq
        self.assertEquals(comp('b1'), m('b1'))
        self.assertEquals(comp('b2'), m('b2'))
        self.assertEquals(comp('b3'), m('b3'))
        self.assertEquals(comp('b4'), m('b4'))
        self.assertEquals(comp('b5'), m('b5'))
        self.assertEquals(comp('b6'), m('b6'))
        self.assertEquals(comp('b7'), m('b7'))
        
    def test_coequalizer_13(self):
        A = Set(['a1','a2', 'a3', 'a4'])
        B = Set(['b1','b2','b3','b4','b5','b6','b7'])
        f = Function(A, B, { 'a1':'b1', 'a2':'b3', 'a3':'b2', 'a4': 'b5' })
        g = Function(A, B, { 'a1':'b2', 'a2':'b4', 'a3':'b4', 'a4': 'b5' })
        part1 = Set(['b1', 'b2', 'b3', 'b4'])
        part2 = Set(['b5'])
        part3 = Set(['b6'])
        part4 = Set(['b7'])
        partition = Set([part1,part2,part3,part4])
        CoEq = coequalizer(f,g)
        self.assertEquals(CoEq.vertex, partition)
        self.assertEquals(CoEq.coeq('b1'), part1)
        self.assertEquals(CoEq.coeq('b2'), part1)
        self.assertEquals(CoEq.coeq('b3'), part1)
        self.assertEquals(CoEq.coeq('b4'), part1)
        self.assertEquals(CoEq.coeq('b5'), part2)
        self.assertEquals(CoEq.coeq('b6'), part3)
        self.assertEquals(CoEq.coeq('b7'), part4)
        M = Set(['x','y'])
        m = Function(B, M, {'b1':'x', 'b2':'x', 'b3':'x', 'b4':'x', 'b5':'y', 'b6':'x', 'b7':'y'})
        comp1 = m * f
        comp2 = m * g
        self.assertEquals(comp1('a1'),comp2('a1'))
        self.assertEquals(comp1('a2'),comp2('a2'))
        self.assertEquals(comp1('a3'),comp2('a3'))
        self.assertEquals(comp1('a4'),comp2('a4'))
        M_cocone = CoCone(M, [m])
        u = CoEq.unique_arrow(M_cocone)
        self.assertEquals(u(part1), 'x')
        self.assertEquals(u(part2), 'y')
        self.assertEquals(u(part3), 'x')
        self.assertEquals(u(part4), 'y')
        comp = u * CoEq.coeq
        self.assertEquals(comp('b1'), m('b1'))
        self.assertEquals(comp('b2'), m('b2'))
        self.assertEquals(comp('b3'), m('b3'))
        self.assertEquals(comp('b4'), m('b4'))
        self.assertEquals(comp('b5'), m('b5'))
        self.assertEquals(comp('b6'), m('b6'))
        self.assertEquals(comp('b7'), m('b7'))
        
    def test_coequalizer_14(self):
        A = Set(['a1','a2', 'a3', 'a4'])
        B = Set(['b1','b2','b3','b4','b5','b6','b7'])
        f = Function(A, B, { 'a1':'b1', 'a2':'b4', 'a3':'b3', 'a4': 'b2' })
        g = Function(A, B, { 'a1':'b2', 'a2':'b5', 'a3':'b3', 'a4': 'b4' })
        part1 = Set(['b1', 'b2', 'b4', 'b5'])
        part2 = Set(['b3'])
        part3 = Set(['b6'])
        part4 = Set(['b7'])
        partition = Set([part1,part2,part3,part4])
        CoEq = coequalizer(f,g)
        self.assertEquals(CoEq.vertex, partition)
        self.assertEquals(CoEq.coeq('b1'), part1)
        self.assertEquals(CoEq.coeq('b2'), part1)
        self.assertEquals(CoEq.coeq('b3'), part2)
        self.assertEquals(CoEq.coeq('b4'), part1)
        self.assertEquals(CoEq.coeq('b5'), part1)
        self.assertEquals(CoEq.coeq('b6'), part3)
        self.assertEquals(CoEq.coeq('b7'), part4)
        M = Set(['x','y'])
        m = Function(B, M, {'b1':'x', 'b2':'x', 'b3':'y', 'b4':'x', 'b5':'x', 'b6':'x', 'b7':'y'})
        comp1 = m * f
        comp2 = m * g
        self.assertEquals(comp1('a1'),comp2('a1'))
        self.assertEquals(comp1('a2'),comp2('a2'))
        self.assertEquals(comp1('a3'),comp2('a3'))
        self.assertEquals(comp1('a4'),comp2('a4'))
        M_cocone = CoCone(M, [m])
        u = CoEq.unique_arrow(M_cocone)
        self.assertEquals(u(part1), 'x')
        self.assertEquals(u(part2), 'y')
        self.assertEquals(u(part3), 'x')
        self.assertEquals(u(part4), 'y')
        comp = u * CoEq.coeq
        self.assertEquals(comp('b1'), m('b1'))
        self.assertEquals(comp('b2'), m('b2'))
        self.assertEquals(comp('b3'), m('b3'))
        self.assertEquals(comp('b4'), m('b4'))
        self.assertEquals(comp('b5'), m('b5'))
        self.assertEquals(comp('b6'), m('b6'))
        self.assertEquals(comp('b7'), m('b7'))
        

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.test_product']
    unittest.main()